@@ -75,12 +75,10 @@ static sigset_t block_sigs;
 /* Pre-calculated kernel packet length */
 static size_t kpkt_len;
 
-/* Does kernel know about SOCK_CLOEXEC and friends */
-static int cloexec_works = 0;
-
 /* Attributes for creating detached and joinable threads */
 pthread_attr_t th_attr;
 pthread_attr_t th_attr_detached;
+size_t detached_thread_stack_size = PTHREAD_STACK_MIN * 144;
 
 struct master_readmap_cond mrc = {
 	PTHREAD_MUTEX_INITIALIZER, PTHREAD_COND_INITIALIZER, 0, NULL, 0, 0, 0, 0};
@@ -89,6 +87,7 @@ struct startup_cond suc = {
 	PTHREAD_MUTEX_INITIALIZER, PTHREAD_COND_INITIALIZER, 0, 0};
 
 pthread_key_t key_thread_stdenv_vars;
+pthread_key_t key_thread_attempt_id = (pthread_key_t) 0L;
 
 #define MAX_OPEN_FILES		10240
 
@@ -98,10 +97,61 @@ static int umount_all(struct autofs_point *ap, int force);
 
 extern struct master *master_list;
 
+/* simple string hash based on public domain sdbm library */
+static unsigned long sdbm_hash(const char *str, unsigned long seed)
+{
+	unsigned long hash = seed;
+	char c;
+
+	while ((c = *str++))
+		hash = c + (hash << 6) + (hash << 16) - hash;
+	return hash;
+}
+
+void set_thread_mount_request_log_id(struct pending_args *mt)
+{
+	char attempt_id_comp[20];
+	unsigned long *attempt_id;
+	int status;
+
+	if (!defaults_get_use_mount_request_log_id())
+		return;
+
+	attempt_id = pthread_getspecific(key_thread_attempt_id);
+	if (attempt_id == NULL) {
+		attempt_id = (unsigned long *) calloc(1, sizeof(unsigned long));
+		if (attempt_id  == NULL)
+			fatal(ENOMEM);
+		snprintf(attempt_id_comp, 20, "%ld", mt->wait_queue_token);
+		*attempt_id = sdbm_hash(attempt_id_comp, 0);
+		snprintf(attempt_id_comp, 20, "%u", mt->pid);
+		*attempt_id = sdbm_hash(attempt_id_comp, *attempt_id);
+		*attempt_id = sdbm_hash(mt->name, *attempt_id);
+		status = pthread_setspecific(key_thread_attempt_id, attempt_id);
+		if (status != 0)
+			fatal(status);
+	}
+}
+
+static int is_remote_fstype(unsigned int fs_type)
+{
+	int ret = 0;
+	switch (fs_type) {
+	case SMB_SUPER_MAGIC:
+	case CIFS_MAGIC_NUMBER:
+	case NCP_SUPER_MAGIC:
+	case NFS_SUPER_MAGIC:
+		ret = 1;
+		break;
+	};
+	return ret;
+}
+
 static int do_mkdir(const char *parent, const char *path, mode_t mode)
 {
 	int status;
-	struct stat st;
+	mode_t mask;
+	struct stat st, root;
 	struct statfs fs;
 
 	/* If path exists we're done */
@@ -114,24 +164,37 @@ static int do_mkdir(const char *parent, const char *path, mode_t mode)
 	}
 
 	/*
-	 * If we're trying to create a directory within an autofs fs
-	 * or the path is contained in a localy mounted fs go ahead.
+	 * We don't want to create the path on a remote file system
+	 * unless it's the root file system.
+	 * An empty parent means it's the root directory and always ok.
 	 */
-	status = -1;
-	if (*parent)
+	if (*parent) {
 		status = statfs(parent, &fs);
-	if ((status != -1 && fs.f_type == (__SWORD_TYPE) AUTOFS_SUPER_MAGIC) ||
-	    contained_in_local_fs(path)) {
-		mode_t mask = umask(0022);
-		int ret = mkdir(path, mode);
-		(void) umask(mask);
-		if (ret == -1) {
-			errno = EACCES;
-			return 0;
+		if (status == -1)
+			goto fail;
+
+		if (is_remote_fstype(fs.f_type)) {
+			status = stat(parent, &st);
+			if (status == -1)
+				goto fail;
+
+			status = stat("/", &root);
+			if (status == -1)
+				goto fail;
+
+			if (st.st_dev != root.st_dev)
+				goto fail;
 		}
-		return 1;
 	}
 
+	mask = umask(0022);
+	status = mkdir(path, mode);
+	(void) umask(mask);
+	if (status == -1)
+		goto fail;
+
+	return 1;
+fail:
 	errno = EACCES;
 	return 0;
 }
@@ -210,7 +273,7 @@ int rmdir_path(struct autofs_point *ap, const char *path, dev_t dev)
 			crit(ap->logopt, "attempt to remove directory from a "
 			     "non-autofs filesystem!");
 			crit(ap->logopt,
-			     "requestor dev == %llu, \"%s\" owner dev == %llu",
+			     "requester dev == %llu, \"%s\" owner dev == %llu",
 			     dev, buf, st.st_dev);
 			return -1;
 		}
@@ -282,17 +345,13 @@ static int walk_tree(const char *base, int (*fn) (struct autofs_point *ap,
 				return -1;
 
 			while (n--) {
-				int ret, size;
-
 				if (strcmp(de[n]->d_name, ".") == 0 ||
 				    strcmp(de[n]->d_name, "..") == 0) {
 					free(de[n]);
 					continue;
 				}
 
-				size = sizeof(buf);
-				ret = cat_path(buf, size, base, de[n]->d_name);
-				if (!ret) {
+				if (!cat_path(buf, sizeof(buf), base, de[n]->d_name)) {
 					do {
 						free(de[n]);
 					} while (n--);
@@ -392,7 +451,7 @@ int count_mounts(struct autofs_point *ap, const char *path, dev_t dev)
 	counter.count = 0;
 	counter.dev = dev;
 	
-	if (walk_tree(path, counter_fn, 0, ap, &counter) == -1)
+	if (walk_tree(path, counter_fn, 1, ap, &counter) == -1)
 		return -1;
 
 	return counter.count;
@@ -548,16 +607,7 @@ static int umount_subtree_mounts(struct autofs_point *ap, const char *path, unsi
 		}
 		list_del(&entry->entries);
 		mounts_mutex_unlock(ap);
-		if (ext_mount_remove(&entry->ext_mount, entry->fs)) {
-			if (umount_ent(ap, entry->fs))
-				debug(ap->logopt,
-				      "failed to umount external mount %s",
-				      entry->fs);
-			else
-				debug(ap->logopt,
-				      "umounted external mount %s",
-				      entry->fs);
-		}
+		umount_amd_ext_mount(ap, entry);
 		free_amd_entry(entry);
 	}
 done:
@@ -574,10 +624,6 @@ int umount_multi(struct autofs_point *ap, const char *path, int incl)
 
 	debug(ap->logopt, "path %s incl %d", path, incl);
 
-	/* If path is a mount it can't be a symlink */
-	if (is_mounted(_PATH_MOUNTED, path, MNTS_ALL))
-		goto real_mount;
-
 	if (lstat(path, &st)) {
 		warn(ap->logopt,
 		     "failed to stat directory or symlink %s", path);
@@ -608,21 +654,11 @@ int umount_multi(struct autofs_point *ap, const char *path, int incl)
 		}
 		list_del(&entry->entries);
 		mounts_mutex_unlock(ap);
-		if (ext_mount_remove(&entry->ext_mount, entry->fs)) {
-			if (umount_ent(ap, entry->fs))
-				debug(ap->logopt,
-				      "failed to umount external mount %s",
-				      entry->fs);
-			else
-				debug(ap->logopt,
-				      "umounted external mount %s",
-				      entry->fs);
-		}
+		umount_amd_ext_mount(ap, entry);
 		free_amd_entry(entry);
 		return 0;
 	}
 
-real_mount:
 	is_autofs_fs = 0;
 	if (master_find_submount(ap, path))
 		is_autofs_fs = 1;
@@ -917,7 +953,7 @@ static int set_log_priority(const char *path, int priority)
 
 	if (priority > LOG_DEBUG || priority < LOG_EMERG) {
 		fprintf(stderr, "Log priority %d is invalid.\n", priority);
-		fprintf(stderr, "Please spcify a number in the range 0-7.\n");
+		fprintf(stderr, "Please specify a number in the range 0-7.\n");
 		return -1;
 	}
 
@@ -938,7 +974,7 @@ static int set_log_priority(const char *path, int priority)
 	 * Specify O_NONBLOCK so that the open will fail if there is no
 	 * daemon reading from the other side of the FIFO.
 	 */
-	fd = open(fifo_name, O_WRONLY|O_NONBLOCK);
+	fd = open_fd(fifo_name, O_WRONLY|O_NONBLOCK);
 	if (fd < 0) {
 		fprintf(stderr, "%s: open of %s failed with %s\n",
 			__FUNCTION__, fifo_name, strerror(errno));
@@ -968,6 +1004,8 @@ static int get_pkt(struct autofs_point *ap, union autofs_v5_packet_union *pkt)
 	struct pollfd fds[3];
 	int pollfds = 3;
 	char buf[MAX_ERR_BUF];
+	size_t read;
+	char *estr;
 
 	fds[0].fd = ap->pipefd;
 	fds[0].events = POLLIN;
@@ -980,7 +1018,6 @@ static int get_pkt(struct autofs_point *ap, union autofs_v5_packet_union *pkt)
 
 	for (;;) {
 		if (poll(fds, pollfds, -1) == -1) {
-			char *estr;
 			if (errno == EINTR)
 				continue;
 			estr = strerror_r(errno, buf, MAX_ERR_BUF);
@@ -999,7 +1036,13 @@ static int get_pkt(struct autofs_point *ap, union autofs_v5_packet_union *pkt)
 
 			state_pipe = ap->state_pipe[0];
 
-			if (fullread(state_pipe, &next_state, read_size)) {
+			read = fullread(state_pipe, &next_state, read_size);
+			if (read) {
+				estr = strerror_r(errno, buf, MAX_ERR_BUF);
+				error(ap->logopt,
+				      "read error on state pipe, "
+				      "read %u, error %s",
+				      read, estr);
 				st_mutex_unlock();
 				continue;
 			}
@@ -1010,8 +1053,17 @@ static int get_pkt(struct autofs_point *ap, union autofs_v5_packet_union *pkt)
 				return -1;
 		}
 
-		if (fds[0].revents & POLLIN)
-			return fullread(ap->pipefd, pkt, kpkt_len);
+		if (fds[0].revents & POLLIN) {
+			read = fullread(ap->pipefd, pkt, kpkt_len);
+			if (read) {
+				estr = strerror_r(errno, buf, MAX_ERR_BUF);
+				error(ap->logopt,
+				      "read error on request pipe, "
+				      "read %u, expected %u error %s",
+				       read, kpkt_len, estr);
+			}
+			return read;
+		}
 
 		if (fds[2].fd != -1 && fds[2].revents & POLLIN) {
 			debug(ap->logopt, "message pending on control fifo.");
@@ -1034,7 +1086,7 @@ int do_expire(struct autofs_point *ap, const char *name, int namelen)
 	}
 
 	if (!len) {
-		crit(ap->logopt, "path to long for buffer");
+		crit(ap->logopt, "path too long for buffer");
 		return 1;
 	}
 
@@ -1363,9 +1415,10 @@ static void *do_read_master(void *arg)
 	return NULL;
 }
 
-static int do_hup_signal(struct master *master, time_t age)
+static int do_hup_signal(struct master *master)
 {
 	unsigned int logopt = master->logopt;
+	time_t age = monotonic_time(NULL);
 	pthread_t thid;
 	int status;
 
@@ -1454,7 +1507,7 @@ static void *statemachine(void *arg)
 			break;
 
 		case SIGHUP:
-			do_hup_signal(master_list, time(NULL));
+			do_hup_signal(master_list);
 			break;
 
 		default:
@@ -1643,6 +1696,100 @@ static void submount_source_unlock_nested(struct autofs_point *ap)
 	master_source_unlock(parent->entry);
 }
 
+int handle_mounts_exit(struct autofs_point *ap)
+{
+	int ret, cur_state;
+
+	/*
+	 * If we're a submount we need to ensure our parent
+	 * doesn't try to mount us again until our shutdown
+	 * is complete and that any outstanding mounts are
+	 * completed before we try to shutdown.
+	 */
+	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
+
+	master_mutex_lock();
+
+	if (!ap->submount)
+		master_source_writelock(ap->entry);
+	else {
+		/*
+		 * If a mount request arrives before the locks are
+		 * aquired just return to ready state.
+		 */
+		ret = submount_source_writelock_nested(ap);
+		if (ret) {
+			warn(ap->logopt,
+			     "can't shutdown submount: mount in progress");
+			/* Return to ST_READY is done immediately */
+			st_add_task(ap, ST_READY);
+			master_mutex_unlock();
+			pthread_setcancelstate(cur_state, NULL);
+			return 0;
+		}
+	}
+
+	if (ap->state != ST_SHUTDOWN) {
+		if (!ap->submount)
+			alarm_add(ap, ap->exp_runfreq);
+		/* Return to ST_READY is done immediately */
+		st_add_task(ap, ST_READY);
+		if (ap->submount)
+			submount_source_unlock_nested(ap);
+		else
+			master_source_unlock(ap->entry);
+		master_mutex_unlock();
+
+		pthread_setcancelstate(cur_state, NULL);
+		return 0;
+	}
+
+	alarm_delete(ap);
+	st_remove_tasks(ap);
+	st_wait_task(ap, ST_ANY, 0);
+
+	/*
+	 * For a direct mount map all mounts have already gone
+	 * by the time we get here and since we only ever
+	 * umount direct mounts at shutdown there is no need
+	 * to check for possible recovery.
+	 */
+	if (ap->type == LKP_DIRECT) {
+		umount_autofs(ap, NULL, 1);
+		handle_mounts_cleanup(ap);
+		return 1;
+	}
+
+	/*
+	 * If umount_autofs returns non-zero it wasn't able
+	 * to complete the umount and has left the mount intact
+	 * so we can continue. This can happen if a lookup
+	 * occurs while we're trying to umount.
+	 */
+	ret = umount_autofs(ap, NULL, 1);
+	if (!ret) {
+		set_indirect_mount_tree_catatonic(ap);
+		handle_mounts_cleanup(ap);
+		return 1;
+	}
+
+	/* Failed shutdown returns to ready */
+	warn(ap->logopt, "can't shutdown: filesystem %s still busy", ap->path);
+	if (!ap->submount)
+		alarm_add(ap, ap->exp_runfreq);
+	/* Return to ST_READY is done immediately */
+	st_add_task(ap, ST_READY);
+	if (ap->submount)
+		submount_source_unlock_nested(ap);
+	else
+		master_source_unlock(ap->entry);
+	master_mutex_unlock();
+
+	pthread_setcancelstate(cur_state, NULL);
+
+	return 0;
+}
+
 void *handle_mounts(void *arg)
 {
 	struct startup_cond *suc;
@@ -1698,99 +1845,24 @@ void *handle_mounts(void *arg)
 
 	pthread_setcancelstate(cancel_state, NULL);
 
-	while (ap->state != ST_SHUTDOWN) {
+	while (1) {
 		if (handle_packet(ap)) {
-			int ret, cur_state;
-
-			/*
-			 * If we're a submount we need to ensure our parent
-			 * doesn't try to mount us again until our shutdown
-			 * is complete and that any outstanding mounts are
-			 * completed before we try to shutdown.
-			 */
-			pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-
-			master_mutex_lock();
-
-			if (ap->submount) {
-				/*
-				 * If a mount request arrives before the locks are
-				 * aquired just return to ready state.
-				 */
-				ret = submount_source_writelock_nested(ap);
-				if (ret) {
-					warn(ap->logopt,
-					     "can't shutdown submount: mount in progress");
-					/* Return to ST_READY is done immediately */
-					st_add_task(ap, ST_READY);
-					master_mutex_unlock();
-					pthread_setcancelstate(cur_state, NULL);
-					continue;
-				}
-			} else
-				master_source_writelock(ap->entry);
-
-			if (ap->state != ST_SHUTDOWN) {
-				if (!ap->submount)
-					alarm_add(ap, ap->exp_runfreq);
-				/* Return to ST_READY is done immediately */
-				st_add_task(ap, ST_READY);
-				if (ap->submount)
-					submount_source_unlock_nested(ap);
-				else
-					master_source_unlock(ap->entry);
-				master_mutex_unlock();
-
-				pthread_setcancelstate(cur_state, NULL);
-				continue;
-			}
-
-			alarm_delete(ap);
-			st_remove_tasks(ap);
-			st_wait_task(ap, ST_ANY, 0);
-
-			/*
-			 * For a direct mount map all mounts have already gone
-			 * by the time we get here and since we only ever
-			 * umount direct mounts at shutdown there is no need
-			 * to check for possible recovery.
-			 */
-			if (ap->type == LKP_DIRECT) {
-				umount_autofs(ap, NULL, 1);
+			if (handle_mounts_exit(ap))
 				break;
-			}
+		}
 
-			/*
-			 * If umount_autofs returns non-zero it wasn't able
-			 * to complete the umount and has left the mount intact
-			 * so we can continue. This can happen if a lookup
-			 * occurs while we're trying to umount.
-			 */
-			ret = umount_autofs(ap, NULL, 1);
-			if (!ret)
+		/* If we get here a packet has been received and handled
+		 * and the autofs mount point has not been shutdown. But
+		 * if the autofs mount point has been set to ST_SHUTDOWN
+		 * we should attempt to perform the shutdown cleanup and
+		 * exit if successful.
+		 */
+		if (ap->state == ST_SHUTDOWN) {
+			if (handle_mounts_exit(ap))
 				break;
-
-			/* Failed shutdown returns to ready */
-			warn(ap->logopt,
-			     "can't shutdown: filesystem %s still busy",
-			     ap->path);
-			if (!ap->submount)
-				alarm_add(ap, ap->exp_runfreq);
-			/* Return to ST_READY is done immediately */
-			st_add_task(ap, ST_READY);
-			if (ap->submount)
-				submount_source_unlock_nested(ap);
-			else
-				master_source_unlock(ap->entry);
-			master_mutex_unlock();
-
-			pthread_setcancelstate(cur_state, NULL);
-
 		}
 	}
 
-	handle_mounts_cleanup(ap);
-
 	return NULL;
 }
 
@@ -1839,7 +1911,6 @@ static void usage(void)
 
 static void show_build_info(void)
 {
-	char buf[2048];
 	int count = 0;
 
 	printf("\nLinux automount version %s\n", version);
@@ -1851,8 +1922,6 @@ static void show_build_info(void)
 
 	printf("\nCompile options:\n  ");
 
-	memset(buf, 0, 2048);
-
 #ifndef ENABLE_MOUNT_LOCKING
 	printf("DISABLE_MOUNT_LOCKING ");
 	count = 22;
@@ -1997,16 +2066,84 @@ static void remove_empty_args(char **argv, int *argc)
 	*argc = j;
 }
 
+static void do_master_list_reset(struct master *master)
+{
+	struct list_head *head, *p;
+
+	master_mutex_lock();
+
+	head = &master->mounts;
+	list_for_each(p, head) {
+		struct master_mapent *entry;
+
+		entry = list_entry(p, struct master_mapent, list);
+
+		if (!list_empty(&entry->list))
+			list_del(&entry->list);
+		master_free_mapent_sources(entry, 1);
+		master_free_mapent(entry);
+	}
+
+	master_mutex_unlock();
+}
+
+static int do_master_read_master(struct master *master, int wait)
+{
+	sigset_t signalset;
+	/* Wait must be at least 1 second */
+	unsigned int retry_wait = 2;
+	unsigned int elapsed = 0;
+	int max_wait = wait;
+	int ret = 0;
+	time_t age;
+
+	sigemptyset(&signalset);
+	sigaddset(&signalset, SIGTERM);
+	sigaddset(&signalset, SIGINT);
+	sigaddset(&signalset, SIGHUP);
+	sigprocmask(SIG_UNBLOCK, &signalset, NULL);
+
+	while (1) {
+		struct timespec t = { retry_wait, 0 };
+
+		do_master_list_reset(master);
+
+		age = monotonic_time(NULL);
+		if (master_read_master(master, age, 0)) {
+			ret = 1;
+			break;
+		}
+
+		if (nanosleep(&t, NULL) == -1)
+			break;
+
+		if (max_wait > 0) {
+			elapsed += retry_wait;
+			if (elapsed >= max_wait) {
+				logmsg("problem reading master map, "
+					"maximum wait exceeded");
+				break;
+			}
+		}
+	}
+
+	sigprocmask(SIG_BLOCK, &signalset, NULL);
+
+	return ret;
+}
+
 int main(int argc, char *argv[])
 {
 	int res, opt, status;
 	int logpri = -1;
 	unsigned ghost, logging, daemon_check;
 	unsigned dumpmaps, foreground, have_global_options;
+	unsigned master_read;
+	int master_wait;
 	time_t timeout;
-	time_t age = time(NULL);
+	time_t age = monotonic_time(NULL);
 	struct rlimit rlim;
-	const char *options = "+hp:t:vmdD:fVrO:l:n:CF";
+	const char *options = "+hp:t:vmdD:fVrO:l:n:CFM";
 	static const struct option long_options[] = {
 		{"help", 0, 0, 'h'},
 		{"pid-file", 1, 0, 'p'},
@@ -2023,6 +2160,7 @@ int main(int argc, char *argv[])
 		{"set-log-priority", 1, 0, 'l'},
 		{"dont-check-daemon", 0, 0, 'C'},
 		{"force", 0, 0, 'F'},
+		{"master-wait", 1, 0, 'M'},
 		{0, 0, 0, 0}
 	};
 
@@ -2043,6 +2181,7 @@ int main(int argc, char *argv[])
 	nfs_mount_uses_string_options = check_nfs_mount_version(&vers, &check);
 
 	kpkt_len = get_kpkt_len();
+	master_wait = defaults_get_master_wait();
 	timeout = defaults_get_timeout();
 	ghost = defaults_get_browse_mode();
 	logging = defaults_get_logging();
@@ -2102,6 +2241,10 @@ int main(int argc, char *argv[])
 			dumpmaps = 1;
 			break;
 
+		case 'M':
+			master_wait = getnumopt(optarg, opt);
+			break;
+
 		case 'O':
 			if (!have_global_options) {
 				global_options = strdup(optarg);
@@ -2331,7 +2474,7 @@ int main(int argc, char *argv[])
 
 #ifdef _POSIX_THREAD_ATTR_STACKSIZE
 	if (pthread_attr_setstacksize(
-			&th_attr_detached, PTHREAD_STACK_MIN*64)) {
+			&th_attr_detached, detached_thread_stack_size)) {
 		logerr("%s: failed to set stack size thread attribute!",
 		       program);
 		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
@@ -2342,6 +2485,17 @@ int main(int argc, char *argv[])
 	}
 #endif
 
+	if (pthread_attr_getstacksize(
+			&th_attr_detached, &detached_thread_stack_size)) {
+		logerr("%s: failed to get detached thread stack size!",
+		       program);
+		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
+		close(start_pipefd[1]);
+		release_flag_file();
+		macro_free_global_table();
+		exit(1);
+	}
+
 	info(logging, "Starting automounter version %s, master map %s",
 		version, master_list->name);
 	info(logging, "using kernel protocol version %d.%02d",
@@ -2360,6 +2514,18 @@ int main(int argc, char *argv[])
 		exit(1);
 	}
 
+	status = pthread_key_create(&key_thread_attempt_id, free);
+	if (status) {
+		logerr("%s: failed to create thread data key for attempt ID!",
+		       program);
+		master_kill(master_list);
+		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
+		close(start_pipefd[1]);
+		release_flag_file();
+		macro_free_global_table();
+		exit(1);
+	}
+
 	init_ioctl_ctl();
 
 	if (!alarm_start_handler()) {
@@ -2393,16 +2559,29 @@ int main(int argc, char *argv[])
 	void *dh_tirpc = dlopen("libtirpc.so", RTLD_NOW);
 	if (!dh_tirpc)
 		dh_tirpc = dlopen("libtirpc.so.1", RTLD_NOW);
+	if (!dh_tirpc)
+		dh_tirpc = dlopen("libtirpc.so.3", RTLD_NOW);
 #endif
 
-	if (!master_read_master(master_list, age, 0)) {
-		master_kill(master_list);
-		*pst_stat = 3;
-		res = write(start_pipefd[1], pst_stat, sizeof(*pst_stat));
-		close(start_pipefd[1]);
-		release_flag_file();
-		macro_free_global_table();
-		exit(3);
+	master_read = master_read_master(master_list, age, 0);
+	if (!master_read) {
+		/*
+		 * Read master map, waiting until it is available, unless
+		 * a signal is received, in which case exit returning an
+		 * error.
+		 */
+		if (!do_master_read_master(master_list, master_wait)) {
+			logmsg("%s: warning: could not read at least one "
+				"map source after waiting, continuing ...",
+				 program);
+			/*
+			 * Failed to read master map, continue with what
+			 * we have anyway.
+			 */
+			do_master_list_reset(master_list);
+			age = time(NULL);
+			master_read_master(master_list, age, 1);
+		}
 	}
 
 	/*
