@@ -31,6 +31,7 @@
 
 #define MODULE_PARSE
 #include "automount.h"
+#include "nsswitch.h"
 
 #define MODPREFIX "parse(amd): "
 
@@ -129,6 +130,11 @@ int parse_init(int argc, const char *const *argv, void **context)
 	return 0;
 }
 
+int parse_reinit(int argc, const char *const *argv, void **context)
+{
+	return 0;
+}
+
 static struct substvar *add_lookup_vars(struct autofs_point *ap,
 					const char *key, int key_len,
 					struct map_source *source,
@@ -239,7 +245,7 @@ static int match_my_name(unsigned int logopt, const char *name, struct substvar
 	}
 
 	if (!v || !v->val) {
-		error(logopt, "error: ${host} not set");
+		error(logopt, MODPREFIX "error: ${host} not set");
 		goto out;
 	}
 
@@ -253,17 +259,17 @@ static int match_my_name(unsigned int logopt, const char *name, struct substvar
 	/* Get host canonical name */
 	ret = getaddrinfo(v->val, NULL, &hints, &cni);
 	if (ret) {
-		error(logopt,
+		error(logopt, MODPREFIX
 		      "hostname lookup failed: %s\n", gai_strerror(ret));
 		goto out;
 	}
 
-	hints.ai_flags = 0;
+	hints.ai_flags = AI_V4MAPPED | AI_ADDRCONFIG | AI_CANONNAME;
 
 	/* Resolve comparison name to its names and compare */
 	ret = getaddrinfo(name, NULL, &hints, &ni);
 	if (ret) {
-		error(logopt,
+		error(logopt, MODPREFIX
 		      "hostname lookup failed: %s\n", gai_strerror(ret));
 		freeaddrinfo(cni);
 		goto out;
@@ -276,10 +282,9 @@ static int match_my_name(unsigned int logopt, const char *name, struct substvar
 				  numeric, sizeof(numeric), NULL, 0,
 				  NI_NUMERICHOST);
 		if (ret) {
-			error(logopt,
+			error(logopt, MODPREFIX
 			      "host address info lookup failed: %s\n",
 			      gai_strerror(ret));
-			freeaddrinfo(cni);
 			goto next;
 		}
 
@@ -287,10 +292,9 @@ static int match_my_name(unsigned int logopt, const char *name, struct substvar
 		ret = getnameinfo(haddr->ai_addr, haddr->ai_addrlen,
 				  host, NI_MAXHOST, NULL, 0, 0);
 		if (ret) {
-			error(logopt,
+			error(logopt, MODPREFIX
 			      "host address info lookup failed: %s\n",
 			      gai_strerror(ret));
-			freeaddrinfo(cni);
 			goto next;
 		}
 
@@ -324,7 +328,8 @@ static int eval_selector(unsigned int logopt,
 	case SEL_FLAG_MACRO:
 		v = macro_findvar(sv, s->sel->name, strlen(s->sel->name));
 		if (!v) {
-			error(logopt, "failed to get selector %s", s->sel->name);
+			error(logopt, MODPREFIX
+			      "failed to get selector %s", s->sel->name);
 			return 0;
 		}
 
@@ -493,7 +498,7 @@ static int eval_selector(unsigned int logopt,
 				else
 					v = macro_findvar(sv, "hostd", 5);
 				if (!v || !*v->val) {
-					error(logopt,
+					error(logopt, MODPREFIX
 					     "failed to get value of ${host}");
 					break;
 				}
@@ -660,7 +665,8 @@ static char *normalize_hostname(unsigned int logopt, const char *host,
 
 		ret = getaddrinfo(host, NULL, &hints, &ni);
 		if (ret) {
-			error(logopt, "hostname lookup failed: %s", gai_strerror(ret));
+			error(logopt, MODPREFIX
+			      "hostname lookup failed: %s", gai_strerror(ret));
 			return NULL;
 		}
 		name = strdup(ni->ai_canonname);
@@ -782,6 +788,35 @@ next:
 		sv = macro_addvar(sv, "remopts", 7, entry->remopts);
 	}
 
+	if (entry->mount) {
+		if (!expand_selectors(ap, entry->mount, &expand, sv)) {
+			free(entry->mount);
+			if (entry->umount)
+				free(entry->umount);
+			entry->mount = NULL;
+			entry->umount = NULL;
+			goto done;
+		}
+		debug(logopt, MODPREFIX
+		      "mount expand(\"%s\") -> %s", entry->mount, expand);
+		free(entry->mount);
+		entry->mount = expand;
+		sv = macro_addvar(sv, "mount", 5, entry->mount);
+	}
+
+	if (entry->umount) {
+		if (!expand_selectors(ap, entry->umount, &expand, sv)) {
+			free(entry->umount);
+			entry->umount = NULL;
+			goto done;
+		}
+		debug(logopt, MODPREFIX
+		      "umount expand(\"%s\") -> %s", entry->umount, expand);
+		free(entry->umount);
+		entry->umount = expand;
+		sv = macro_addvar(sv, "umount", 5, entry->umount);
+	}
+done:
 	return sv;
 }
 
@@ -926,7 +961,7 @@ static int do_auto_mount(struct autofs_point *ap, const char *name,
 	}
 
 	return do_mount(ap, ap->path,
-			name, strlen(name), target, "autofs", NULL);
+			name, strlen(name), target, "autofs", entry->opts);
 }
 
 static int do_link_mount(struct autofs_point *ap, const char *name,
@@ -974,8 +1009,7 @@ symlink:
 
 	if (entry->sublink) {
 		/* failed to complete sublink mount */
-		if (ext_mount_remove(&entry->ext_mount, entry->fs))
-			umount_ent(ap, entry->fs);
+		umount_amd_ext_mount(ap, entry);
 	}
 out:
 	return ret;
@@ -1129,6 +1163,7 @@ static int do_host_mount(struct autofs_point *ap, const char *name,
 	struct mapent *me;
 	const char *argv[2];
 	const char **pargv = NULL;
+	int status;
 	int argc = 0;
 	int ret = 1;
 
@@ -1170,8 +1205,8 @@ static int do_host_mount(struct autofs_point *ap, const char *name,
 	}
 
 	instance_mutex_lock();
-	lookup = open_lookup("hosts", MODPREFIX, NULL, argc, pargv);
-	if (!lookup) {
+	status = open_lookup("hosts", MODPREFIX, NULL, argc, pargv, &lookup);
+	if (status != NSS_STATUS_SUCCESS) {
 		debug(ap->logopt, "open lookup module hosts failed");
 		instance_mutex_unlock();
 		goto out;
@@ -1181,7 +1216,7 @@ static int do_host_mount(struct autofs_point *ap, const char *name,
 					 "hosts", "sun", argc, pargv);
 	if (!instance) {
 		instance = master_add_source_instance(source,
-				 "hosts", "sun", time(NULL), argc, pargv);
+				 "hosts", "sun", monotonic_time(NULL), argc, pargv);
 		if (!instance) {
 			error(ap->logopt, MODPREFIX
 			     "failed to create source instance for hosts map");
@@ -1216,6 +1251,91 @@ out:
 	return ret;
 }
 
+static int do_program_mount(struct autofs_point *ap,
+			    struct amd_entry *entry, const char *name)
+{
+	char *prog, *str;
+	char **argv;
+	int argc = -1;
+	int rv = 1;
+
+	str = strdup(entry->mount);
+	if (!str)
+		goto out;
+
+	prog = NULL;
+	argv = NULL;
+
+	argc = construct_argv(str, &prog, &argv);
+	if (argc == -1) {
+		error(ap->logopt, MODPREFIX
+		      "%s: error creating mount arguments", entry->type);
+		free(str);
+		goto out;
+	}
+
+	/* The am-utils documentation doesn't actually say that the
+	 * mount (and umount, if given) command need to use ${fs} as
+	 * the mount point in the command.
+	 *
+	 * For program mounts there's no way to know what the mount
+	 * point is so ${fs} must be used in the mount (and umount,
+	 * if given) in order to create the mount point directory
+	 * before executing the mount command and removing it at
+	 * umount.
+	 */
+	if (ext_mount_inuse(entry->fs)) {
+		rv = 0;
+		ext_mount_add(&entry->ext_mount, entry->fs, 1);
+	} else {
+		rv = mkdir_path(entry->fs, 0555);
+		if (rv && errno != EEXIST) {
+			char buf[MAX_ERR_BUF];
+			char *estr;
+
+			estr = strerror_r(errno, buf, MAX_ERR_BUF);
+			error(ap->logopt,
+			      MODPREFIX "%s: mkdir_path %s failed: %s",
+			      entry->type, entry->fs, estr);
+			goto do_free;
+		}
+
+		rv = spawnv(ap->logopt, prog, (const char * const *) argv);
+		if (WIFEXITED(rv) && !WEXITSTATUS(rv)) {
+			rv = 0;
+			ext_mount_add(&entry->ext_mount, entry->fs, 1);
+			debug(ap->logopt, MODPREFIX
+			      "%s: mounted %s", entry->type, entry->fs);
+		} else {
+			if (!ext_mount_inuse(entry->fs))
+				rmdir_path(ap, entry->fs, ap->dev);
+			error(ap->logopt, MODPREFIX
+			     "%s: failed to mount using: %s",
+			     entry->type, entry->mount);
+		}
+	}
+do_free:
+	free_argv(argc, (const char **) argv);
+	free(str);
+
+	if (rv)
+		goto out;
+
+	rv = do_link_mount(ap, name, entry, 0);
+	if (!rv)
+		goto out;
+
+	if (umount_amd_ext_mount(ap, entry)) {
+		if (!ext_mount_inuse(entry->fs))
+			rmdir_path(ap, entry->fs, ap->dev);
+		debug(ap->logopt, MODPREFIX
+		      "%s: failed to umount external mount at %s",
+		      entry->type, entry->fs);
+	}
+out:
+	return rv;
+}
+
 static unsigned int validate_auto_options(unsigned int logopt,
 					  struct amd_entry *entry)
 {
@@ -1295,7 +1415,7 @@ static unsigned int validate_generic_options(unsigned int logopt,
 			 * Can't use entry->type as the mount type to reprot
 			 * the error since entry->type == "bind" not "lofs".
 			 */
-			error(logopt, "lofs: mount device not given");
+			error(logopt, MODPREFIX "lofs: mount device not given");
 			return 0;
 		} else if (!*entry->rfs)
 			return 0;
@@ -1342,6 +1462,29 @@ static unsigned int validate_host_options(unsigned int logopt,
 	return 1;
 }
 
+static unsigned int validate_program_options(unsigned int logopt,
+					     struct amd_entry *entry)
+{
+	/*
+	 * entry->mount will be NULL if there is a problem expanding
+	 * ${} macros in expandamdent().
+	 */
+	if (!entry->mount) {
+		error(logopt, MODPREFIX
+		      "%s: mount program invalid or not set", entry->type);
+		return 0;
+	}
+
+	if (!entry->fs || !*entry->fs) {
+		error(logopt, MODPREFIX
+		    "%s: ${fs} must be used as the mount point but is not set",
+		    entry->type);
+		return 0;
+	}
+
+	return 1;
+}
+
 static int amd_mount(struct autofs_point *ap, const char *name,
 		     struct amd_entry *entry, struct map_source *source,
 		     struct substvar *sv, unsigned int flags,
@@ -1407,6 +1550,12 @@ static int amd_mount(struct autofs_point *ap, const char *name,
 		ret = do_host_mount(ap, name, entry, source, flags);
 		break;
 
+	case AMD_MOUNT_TYPE_PROGRAM:
+		if (!validate_program_options(ap->logopt, entry))
+			return 1;
+		ret = do_program_mount(ap, entry, name);
+		break;
+
 	default:
 		info(ap->logopt,
 		     MODPREFIX "unknown file system type %x", fstype);
@@ -1568,7 +1717,7 @@ static int match_selectors(unsigned int logopt,
 
 	/* No selectors, always match */
 	if (!s) {
-		debug(logopt, "no selectors found in location");
+		debug(logopt, MODPREFIX "no selectors found in location");
 		return 1;
 	}
 
@@ -1823,55 +1972,24 @@ out:
 	return make_default_entry(ap, sv);
 }
 
-int parse_mount(struct autofs_point *ap, const char *name,
-		int name_len, const char *mapent, void *context)
+static struct amd_entry *setup_defaults(struct autofs_point *ap,
+					const char *name, int name_len,
+					struct map_source *source,
+					struct substvar **sv)
 {
-	struct parse_context *ctxt = (struct parse_context *) context;
-	unsigned int flags = conf_amd_get_flags(ap->path);
-	struct substvar *sv = NULL;
-	struct map_source *source;
+	struct amd_entry *defaults_entry;
 	struct mapent_cache *mc;
 	struct mapent *me;
-	unsigned int at_least_one;
-	struct list_head entries, *p, *head;
-	struct amd_entry *defaults_entry;
-	struct amd_entry *cur_defaults;
+	struct substvar *nsv;
 	char *defaults;
-	char *pmapent;
-	int len, rv = 1;
-	int cur_state;
-	int ret;
-
-	source = ap->entry->current;
-	ap->entry->current = NULL;
-	master_source_current_signal(ap->entry);
 
 	mc = source->mc;
+	defaults = NULL;
+	defaults_entry = NULL;
 
-	if (!mapent) {
-		warn(ap->logopt, MODPREFIX "error: empty map entry");
-		return 1;
-	}
-
-	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
-
-	sv = add_lookup_vars(ap, name, name_len, source, sv);
-	if (!sv) {
-		macro_free_table(sv);
-		pthread_setcancelstate(cur_state, NULL);
-		return 1;
-	}
-
-	len = expand_selectors(ap, mapent, &pmapent, sv);
-	if (!len) {
-		macro_free_table(sv);
-		pthread_setcancelstate(cur_state, NULL);
-		return 1;
-	}
-
-	pthread_setcancelstate(cur_state, NULL);
-
-	debug(ap->logopt, MODPREFIX "expanded mapent: %s", pmapent);
+	nsv = add_lookup_vars(ap, name, name_len, source, NULL);
+	if (!nsv)
+		goto done;
 
 	defaults = conf_amd_get_map_defaults(ap->path);
 	if (defaults) {
@@ -1886,32 +2004,69 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			char buf[MAX_ERR_BUF];
 			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
 			error(ap->logopt, MODPREFIX "malloc: %s", estr);
+			macro_free_table(nsv);
+			nsv = NULL;
+			goto done;
 		}
 	}
 
-	defaults_entry = get_defaults_entry(ap, defaults, sv);
+	defaults_entry = get_defaults_entry(ap, defaults, nsv);
 	if (!defaults_entry) {
 		error(ap->logopt, MODPREFIX "failed to get a defaults entry");
-		if (defaults)
-			free(defaults);
-		free(pmapent);
-		macro_free_table(sv);
-		return 1;
+		macro_free_table(nsv);
+		nsv = NULL;
 	}
+done:
 	if (defaults)
 		free(defaults);
+	if (*sv)
+		macro_free_table(*sv);
+	*sv = nsv;
+
+	return defaults_entry;
+}
+
+int parse_mount(struct autofs_point *ap, const char *name,
+		int name_len, const char *mapent, void *context)
+{
+	struct parse_context *ctxt = (struct parse_context *) context;
+	unsigned int flags = conf_amd_get_flags(ap->path);
+	struct substvar *sv = NULL;
+	struct map_source *source;
+	unsigned int at_least_one;
+	struct list_head entries, *p, *head;
+	struct amd_entry *defaults_entry;
+	struct amd_entry *cur_defaults;
+	int rv = 1;
+	int cur_state;
+	int ret;
+
+	source = ap->entry->current;
+	ap->entry->current = NULL;
+	master_source_current_signal(ap->entry);
+
+	if (!mapent) {
+		warn(ap->logopt, MODPREFIX "error: empty map entry");
+		return 1;
+	}
 
 	INIT_LIST_HEAD(&entries);
 
-	ret = amd_parse_list(ap, pmapent, &entries, &sv);
-	if (ret) {
-		error(ap->logopt,
-		      MODPREFIX "failed to parse entry: %s", pmapent);
-		free(pmapent);
+	pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, &cur_state);
+
+	defaults_entry = setup_defaults(ap, name, name_len, source, &sv);
+	if (!defaults_entry) {
+		error(ap->logopt, MODPREFIX
+		      "failed to setup defaults entry");
 		goto done;
 	}
 
-	free(pmapent);
+	ret = amd_parse_list(ap, mapent, &entries, &sv);
+	if (ret) {
+		error(ap->logopt,
+		      MODPREFIX "failed to parse entry: %s", mapent);
+		goto done;
+	}
 
 	if (list_empty(&entries)) {
 		error(ap->logopt, MODPREFIX "no location found after parse");
@@ -1936,9 +2091,20 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			free_amd_entry(cur_defaults);
 			list_del_init(&this->list);
 			cur_defaults = this;
+			update_with_defaults(defaults_entry, cur_defaults, sv);
 			continue;
 		} else if (this->flags & AMD_DEFAULTS_RESET) {
-			struct amd_entry *new;
+			struct amd_entry *nd, *new;
+			struct substvar *nsv = NULL;
+
+			nd = setup_defaults(ap, name, name_len, source, &nsv);
+			if (nd) {
+				free_amd_entry(defaults_entry);
+				defaults_entry = nd;
+				macro_free_table(sv);
+				sv = nsv;
+			}
+
 			new = dup_defaults_entry(defaults_entry);
 			if (new) {
 				free_amd_entry(cur_defaults);
@@ -1949,6 +2115,9 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			continue;
 		}
 
+		debug(ap->logopt, "expand defaults entry");
+		sv = expand_entry(ap, cur_defaults, flags, sv);
+
 		if (this->flags & AMD_ENTRY_CUT && at_least_one) {
 			info(ap->logopt, MODPREFIX
 			     "at least one entry tried before cut selector, "
@@ -1961,6 +2130,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 
 		at_least_one = 1;
 
+		debug(ap->logopt, "expand mount entry");
 		update_with_defaults(cur_defaults, this, sv);
 		sv = expand_entry(ap, this, flags, sv);
 		sv = merge_entry_options(ap, this, sv);
@@ -1993,12 +2163,15 @@ int parse_mount(struct autofs_point *ap, const char *name,
 	free_amd_entry(cur_defaults);
 
 	if (rv)
-		debug(ap->logopt, "no more locations to try, returning fail");
+		debug(ap->logopt, MODPREFIX
+		      "no more locations to try, returning fail");
 done:
 	free_amd_entry_list(&entries);
 	free_amd_entry(defaults_entry);
 	macro_free_table(sv);
 
+	pthread_setcancelstate(cur_state, NULL);
+
 	return rv;
 }
 
